@@ -22,6 +22,10 @@
 #include <sys/wait.h>
 #include <unistd.h>
 
+#ifndef MAX_FORKS
+# define MAX_FORKS	200
+#endif
+
 /*
  * tmpreaper.c -- remove files in a directory, but do it carefully
  */
@@ -34,9 +38,9 @@
 #define FLAGS_MTIME		(1 << 1)
 #define FLAGS_MTIME_P(fl)	((fl) & FLAGS_MTIME)
 
-/* `--mtime-mt-dir': rm empty directories based on mtime */
-#define FLAGS_MTIME_MT_DIR	(1 << 2)
-#define FLAGS_MTIME_MT_DIR_P(fl) ((fl) & FLAGS_MTIME_MT)
+/* `--mtime-dir': rm empty directories based on mtime */
+#define FLAGS_MTIME_DIR		(1 << 2)
+#define FLAGS_MTIME_DIR_P(fl)	((fl) & FLAGS_MTIME_DIR)
 
 /* `--symlinks': also rm symlinks like files and directories */
 #define FLAGS_SYMLINKS		(1 << 3)
@@ -219,7 +223,7 @@ dir_empty_p (const char * dirname)
 }
 
 int
-cleanupDirectory (const char	   * dirname,
+cleanupDirectory (const char	   * dirname ,
 		  const unsigned int killTime,
 		  const int	     flags)
 {
@@ -228,6 +232,7 @@ cleanupDirectory (const char	   * dirname,
     struct stat	    sb, here;
     int		    status, pid, skip = 0;
     int		    i;
+    static int	    fork_count = 0;
 
     if (FLAGS_TEST_P (flags)) {
 	message (LOG_VERBOSE,
@@ -239,6 +244,30 @@ cleanupDirectory (const char	   * dirname,
 		"Cleaning up directory `%s'.\n",
 		dirname);
     }
+    if (fork_count > MAX_FORKS) {
+	char	path[PATH_MAX], *p;
+	int	l;
+	/* the admin will want to know where the problem lies */
+	if (getcwd(path, PATH_MAX) < 0) {
+	    /* too long even for PATH_MAX, show something anyway */
+	    p = "..../";
+	}
+	else {
+	    l = strlen(path);
+	    if (l > 60) {
+		    /* remove middle part of the path */
+		    p = path + l - 30;
+		    strcpy(&path[30], " ... ");
+		    memmove(&path[35], p, 30);
+		    path[65] = 0;
+	    }
+	    p = path;
+	}
+	message(LOG_FATAL,
+		"Too deep directory nesting detected while trying to clean up\n`%s/%s',\npossible attack?\n",
+		p, dirname);
+	exit(2);
+    }
 
     /*
      * Do everything in a child process so we don't have to chdir(".."),
@@ -258,6 +287,7 @@ cleanupDirectory (const char	   * dirname,
     if (! (pid = fork())) {
 	    if (safe_chdir (dirname))
 		    return 1;
+	    fork_count++;
 	    if (! (dir = opendir ("."))) {
 		message (LOG_FATAL,
 			 "Opening directory, `%s' as `.' : %s\n",
@@ -281,8 +311,17 @@ cleanupDirectory (const char	   * dirname,
 		if (!ent)
 		    break;
 
-		if (((ent->d_name[0] == '.') && (ent->d_name[1] == '\0')) ||
-		    ((ent->d_name[1] == '.') && (ent->d_name[2] == '\0')))
+		/*
+		 * if name starts with dot, and
+		 *    is 1 byte long, or
+		 *    has a second dot and is 2 bytes long,
+		 *   skip it.
+		 */
+		if ((ent->d_name[0] == '.') && (
+			 (ent->d_name[1] == '\0') ||
+			 ((ent->d_name[1] == '.') && (ent->d_name[2] == '\0'))
+		    )
+		   )
 		    continue;
 
 		message (LOG_DEBUG,
@@ -332,7 +371,9 @@ cleanupDirectory (const char	   * dirname,
 		}
 
 		/* Decide whether to remove the file or not */
-		if (FLAGS_MTIME_P (flags)) {
+		/* check for mtime on directory instead of atime if requested */
+		if ( FLAGS_MTIME_P(flags) ||
+		    (FLAGS_MTIME_DIR_P(flags) && S_ISDIR(sb.st_mode))) {
 		    if (sb.st_mtime >= killTime)
 			continue;
 		}
@@ -393,9 +434,12 @@ cleanupDirectory (const char	   * dirname,
 	    closedir (dir);
 	    
 	    exit (0);
-	}
+    }
 
-    waitpid (pid, &status, 0);
+    if (waitpid(pid, &status, 0) < 0) {
+	message(LOG_ERROR, "waitpid failed: %s\n", strerror(errno));
+	return 1;
+    }
 
     if (WIFEXITED (status))
 	return WEXITSTATUS (status);
@@ -420,8 +464,9 @@ usage(void)
 {
     printCopyright ();
     fprintf (stderr, "\n"
-	     "tmpreaper [-htvfmsa] [--help] [--test] [--verbose] [--force] [--mtime]\n"
-	     "[--symlinks] [--all] [[--protect '<shell_pattern>']...] <time> <dirs>...\n"
+	     "tmpreaper [-htvfmMsa] [--help] [--test] [--verbose] [--force]\n"
+	     "          [--mtime] [--mtime-dir] [--symlinks] [--all]\n"
+	     "          [[--protect '<shell_pattern>']...] <time> <dirs>...\n"
 	     "<time> is time since a file in a <dir> was last accessed.\n"
 	     "It defaults to hours, or you may suffix with `s', `m', `h', or `d'.\n\n");
     exit (1);
@@ -455,14 +500,15 @@ main (int     argc,
     char **p = protect_argv;
 
     struct option const options[] = {
-	{ "help",    no_argument,	NULL, 'h' },
-	{ "test",    no_argument,	NULL, 't' },
-	{ "verbose", no_argument,	NULL, 'v' },
-	{ "force",   no_argument,	NULL, 'f' },
-	{ "mtime",   no_argument,	NULL, 'm' },
-	{ "symlinks",no_argument,	NULL, 's' },
-	{ "all",     no_argument,	NULL, 'a' },
-	{ "protect", required_argument, NULL, 'p' },
+	{ "help",	no_argument,		NULL, 'h' },
+	{ "test",	no_argument,		NULL, 't' },
+	{ "verbose",	no_argument,		NULL, 'v' },
+	{ "force",	no_argument,		NULL, 'f' },
+	{ "mtime",	no_argument,		NULL, 'm' },
+	{ "mtime-dirs",	no_argument,		NULL, 'M' },
+	{ "symlinks",	no_argument,		NULL, 's' },
+	{ "all",	no_argument,		NULL, 'a' },
+	{ "protect",	required_argument,	NULL, 'p' },
 	{ 0, 0, 0, 0 }
     };
 
@@ -472,7 +518,7 @@ main (int     argc,
     while (1) {
 	long_index = 0;
 
-	arg = getopt_long (argc, argv, "htvfmsa:", options, &long_index);
+	arg = getopt_long (argc, argv, "htvfmMsa:", options, &long_index);
 	if (arg == -1)
 	    break;
 
@@ -498,6 +544,10 @@ main (int     argc,
 	    flags |= FLAGS_MTIME;
 	    break;
 
+	case 'M':
+	    flags |= FLAGS_MTIME_DIR;
+	    break;
+
 	case 's':
 	    flags |= FLAGS_SYMLINKS;
 	    break;
